@@ -1,4 +1,4 @@
-// Copyright 2011-2020 David Robillard <d@drobilla.net>
+// Copyright 2011-2023 David Robillard <d@drobilla.net>
 // SPDX-License-Identifier: ISC
 
 #include "string_utils.h"
@@ -17,8 +17,8 @@ serd_uri_to_path(const uint8_t* uri)
 {
   const uint8_t* path = uri;
   if (!is_windows_path(uri) && serd_uri_string_has_scheme(uri)) {
-    if (strncmp((const char*)uri, "file:", 5)) {
-      fprintf(stderr, "Non-file URI `%s'\n", uri);
+    if (!!strncmp((const char*)uri, "file:", 5)) {
+      fprintf(stderr, "Non-file URI '%s'\n", uri);
       return NULL;
     }
 
@@ -27,7 +27,7 @@ serd_uri_to_path(const uint8_t* uri)
     } else if (!strncmp((const char*)uri, "file://", 7)) {
       path = uri + 7;
     } else {
-      fprintf(stderr, "Invalid file URI `%s'\n", uri);
+      fprintf(stderr, "Invalid file URI '%s'\n", uri);
       return NULL;
     }
 
@@ -72,8 +72,9 @@ serd_file_uri_parse(const uint8_t* const uri, uint8_t** const hostname)
         serd_chunk_sink("%", 1, &chunk);
         ++s;
       } else if (is_hexdig(*(s + 1)) && is_hexdig(*(s + 2))) {
-        const uint8_t code[3] = {*(s + 1), *(s + 2), 0};
-        const uint8_t c       = (uint8_t)strtoul((const char*)code, NULL, 16);
+        const uint8_t hi = hex_digit_value(s[1]);
+        const uint8_t lo = hex_digit_value(s[2]);
+        const char    c  = (char)((hi << 4U) | lo);
         serd_chunk_sink(&c, 1, &chunk);
         s += 2;
       } else {
@@ -236,64 +237,26 @@ remove_dot_segments(const uint8_t* const path,
                     const size_t         len,
                     size_t* const        up)
 {
-  const uint8_t*       begin = path;
-  const uint8_t* const end   = path + len;
-
   *up = 0;
-  while (begin < end) {
-    switch (begin[0]) {
-    case '.':
-      switch (begin[1]) {
-      case '/':
-        begin += 2; // Chop leading "./"
-        break;
-      case '.':
-        switch (begin[2]) {
-        case '\0':
-          ++*up;
-          begin += 2; // Chop input ".."
-          break;
-        case '/':
-          ++*up;
-          begin += 3; // Chop leading "../"
-          break;
-        default:
-          return begin;
-        }
-        break;
-      case '\0':
-        return ++begin; // Chop input "."
-      default:
-        return begin;
-      }
-      break;
 
-    case '/':
-      switch (begin[1]) {
-      case '.':
-        switch (begin[2]) {
-        case '/':
-          begin += 2; // Leading "/./" => "/"
-          break;
-        case '.':
-          switch (begin[3]) {
-          case '/':
-            ++*up;
-            begin += 3; // Leading "/../" => "/"
-          }
-          break;
-        default:
-          return begin;
-        }
-      }
-      return begin;
-
-    default:
-      return begin; // Finished chopping dot components
+  for (size_t i = 0; i < len;) {
+    const char* const p = (char*)path + i;
+    if (!strcmp(p, ".")) {
+      ++i; // Chop input "."
+    } else if (!strcmp(p, "..")) {
+      ++*up;
+      i += 2; // Chop input ".."
+    } else if (!strncmp(p, "./", 2) || !strncmp(p, "/./", 3)) {
+      i += 2; // Chop leading "./", or replace leading "/./" with "/"
+    } else if (!strncmp(p, "../", 3) || !strncmp(p, "/../", 4)) {
+      ++*up;
+      i += 3; // Chop leading "../", or replace "/../" with "/"
+    } else {
+      return (uint8_t*)p;
     }
   }
 
-  return begin;
+  return path + len;
 }
 
 /// Merge `base` and `path` in-place
@@ -428,10 +391,6 @@ write_rel_path(SerdSink             sink,
     len += sink("../", 3, stream);
   }
 
-  if (last_shared_sep == 0 && up == 0) {
-    len += sink("/", 1, stream);
-  }
-
   // Write suffix
   return len + write_path_tail(sink, stream, uri, last_shared_sep + 1);
 }
@@ -468,8 +427,12 @@ serd_uri_serialise_relative(const SerdURI* const uri,
     if (uri->authority.buf) {
       len += sink("//", 2, stream);
       len += sink(uri->authority.buf, uri->authority.len, stream);
-      if (uri->authority.len > 0 &&
-          uri->authority.buf[uri->authority.len - 1] != '/' &&
+
+      const bool authority_ends_with_slash =
+        (uri->authority.len > 0 &&
+         uri->authority.buf[uri->authority.len - 1] == '/');
+
+      if (!authority_ends_with_slash &&
           serd_uri_path_starts_without_slash(uri)) {
         // Special case: ensure path begins with a slash
         // https://tools.ietf.org/html/rfc3986#section-3.2
@@ -485,7 +448,7 @@ serd_uri_serialise_relative(const SerdURI* const uri,
   }
 
   if (uri->fragment.buf) {
-    // Note uri->fragment.buf includes the leading `#'
+    // Note uri->fragment.buf includes the leading '#'
     len += sink(uri->fragment.buf, uri->fragment.len, stream);
   }
 
